# OMR - Optical mark regognition
# Copyright (C) 2019 Joppe Blondel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import cv2
import numpy as np
import math

SQUARE_TOLERANCE = 0.2
PLACE_TOLERANCE = 5
DIV_TOLERANCE = 0.1
A4_SCALE = 1.41

def has_square_parent(hierarchy, squares, parent):
    if hierarchy[parent][3] == -1:
        return False
    if hierarchy[parent][3] in squares:
        return True
    return has_square_parent(hierarchy, squares, hierarchy[parent][3])

def count_children(hierarchy, parent, inner=False):
    if parent == -1:
        return 0
    elif not inner:
        return count_children(hierarchy, hierarchy[parent][2], True)
    return 1 + count_children(hierarchy, hierarchy[parent][0], True) + count_children(hierarchy, hierarchy[parent][2], True)

def get_angle(p1, p2):
    x_diff = p2[0] - p1[0]
    y_diff = p2[1] - p1[1]
    return math.degrees(math.atan2(y_diff, x_diff))

def prepare_image_create(f):
    img = cv2.imread(f)
    img_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    ret, img_th = cv2.threshold(img_gray, 127, 255, 0)
    i_cont, i_hier = cv2.findContours(img_th, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    height, width = img.shape[:2]

    # Get corner squares
    squares = []
    square_indices = []
    square_midX = []
    square_midY = []
    black = 0
    i = 0
    j = 0
    for c in i_cont:
        peri = cv2.arcLength(c, True)
        area = cv2.contourArea(c)
        approx = cv2.approxPolyDP(c, 0.03* peri, True)
        if len(approx)==4:
            if width*height*(500/2180952)<area<width*height*(10000/2180952):
                approx_area = math.fabs((peri/4) ** 2)
                if 1-SQUARE_TOLERANCE<approx_area/area<SQUARE_TOLERANCE+1 and has_square_parent(i_hier[0], square_indices, i) is False:

                    # Check if right lower corner
                    if count_children(i_hier[0], i) < 2:
                        black = j
                        #cv2.drawContours(img, [c], 0, (255, 0, 0), 8)
                    #else:
                        #cv2.drawContours(img, [c], 0, (0, 0, 255), 8)

                    squares.append(approx)
                    square_indices.append(i)

                    j += 1

                    # Find center
                    M = cv2.moments(c)
                    cX = int(M["m10"] / M["m00"])
                    cY = int(M["m01"] / M["m00"])
                    square_midX.append(cX)
                    square_midY.append(cY)
                    #cv2.circle(img, (cX, cY), 10, (0, 255, 255), -1)
        i += 1

    if j!=4:
        print("Found more or less than 4 corners: " + str(j))

    # Right bottom corner
    sq1 = np.array([square_midX[black], square_midY[black]])
    i_sq1 = black

    # Find smallest biggest and mid
    i_max = 0
    max = 0
    i_min = 0
    min = 0
    i_mid = 0
    for i in range(0,4):
        if i==i_sq1:
            continue
        p = np.array([square_midX[i], square_midY[i]])
        dist = np.linalg.norm(sq1-p)
        if dist>max:
            max = dist
            i_max = i
        if dist<min or min==0:
            min = dist
            i_min = i

    for i in range(0, 4):
        if i!=i_max and i!=i_min and i!=i_sq1:
            i_mid = i
            break

    i_sq2 = i_max
    i_sq3 = i_min
    i_sq4 = i_mid

    sq2 = np.array([square_midX[i_sq2], square_midY[i_sq2]])
    sq3 = np.array([square_midX[i_sq3], square_midY[i_sq3]])
    sq4 = np.array([square_midX[i_sq4], square_midY[i_sq4]])

    # Write numbers on image
    #font                   = cv2.FONT_HERSHEY_SIMPLEX
    #fontScale              = 5
    #fontColor              = (0,255,0)
    #lineType               = 10
    #cv2.putText(img,'1', (sq1[0], sq1[1]), font, fontScale, fontColor, lineType)
    #cv2.putText(img,'2', (sq2[0], sq2[1]), font, fontScale, fontColor, lineType)
    #cv2.putText(img,'3', (sq3[0], sq3[1]), font, fontScale, fontColor, lineType)
    #cv2.putText(img,'4', (sq4[0], sq4[1]), font, fontScale, fontColor, lineType)

    src = np.array([
        sq2,
        sq4,
        sq1,
        sq3], dtype="float32")
    w = width
    h = height
    if w>h:
        w=int(h/A4_SCALE)
    else:
        h=int(w*A4_SCALE)
    dst = np.array([
        [0, 0],
        [w, 0],
        [w, h],
        [0, h]], dtype = "float32")

    transform = cv2.getPerspectiveTransform(src, dst)
    img = cv2.warpPerspective(img, transform, (width, height))
    img = img[0:h, 0:w]
    img_gray = cv2.warpPerspective(img_gray, transform, (width, height))
    img_gray = img_gray[0:h, 0:w]
    return transform, img, img_gray

def prepare_image(f, debug):
    img = cv2.imread(f)
    img_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    ret, img_th = cv2.threshold(img_gray, 127, 255, 0)
    i_cont, i_hier = cv2.findContours(img_th, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    height, width = img.shape[:2]

    # Get corner squares
    squares = []
    square_indices = []
    square_midX = []
    square_midY = []
    black = 0
    i = 0
    j = 0
    for c in i_cont:
        peri = cv2.arcLength(c, True)
        area = cv2.contourArea(c)
        approx = cv2.approxPolyDP(c, 0.03* peri, True)
        if len(approx)==4:
            if width*height*(500/2180952)<area<width*height*(10000/2180952):
                approx_area = math.fabs((peri/4) ** 2)
                if 1-SQUARE_TOLERANCE<approx_area/area<SQUARE_TOLERANCE+1 and has_square_parent(i_hier[0], square_indices, i) is False:

                    # Check if right lower corner
                    if count_children(i_hier[0], i) < 2:
                        black = j
                        cv2.drawContours(img, [c], 0, (255, 0, 0), 8)
                    else:
                        cv2.drawContours(img, [c], 0, (0, 0, 255), 8)

                    squares.append(approx)
                    square_indices.append(i)

                    j += 1

                    # Find center
                    M = cv2.moments(c)
                    cX = int(M["m10"] / M["m00"])
                    cY = int(M["m01"] / M["m00"])
                    square_midX.append(cX)
                    square_midY.append(cY)
                    cv2.circle(img, (cX, cY), 10, (0, 255, 255), -1)
                else:
                    cv2.drawContours(img, [c], 0, (0, 255, 0), 4)
        i += 1

    if debug:
        cv2.imwrite('conrners.jpg', img)

    if j!=4:
        print("Found more or less than 4 corners: " + str(j))

    # Right bottom corner
    sq1 = np.array([square_midX[black], square_midY[black]])
    i_sq1 = black

    # Find smallest biggest and mid
    i_max = 0
    max = 0
    i_min = 0
    min = 0
    i_mid = 0
    for i in range(0,4):
        if i==i_sq1:
            continue
        p = np.array([square_midX[i], square_midY[i]])
        dist = np.linalg.norm(sq1-p)
        if dist>max:
            max = dist
            i_max = i
        if dist<min or min==0:
            min = dist
            i_min = i

    for i in range(0, 4):
        if i!=i_max and i!=i_min and i!=i_sq1:
            i_mid = i
            break

    i_sq2 = i_max
    i_sq3 = i_min
    i_sq4 = i_mid

    sq2 = np.array([square_midX[i_sq2], square_midY[i_sq2]])
    sq3 = np.array([square_midX[i_sq3], square_midY[i_sq3]])
    sq4 = np.array([square_midX[i_sq4], square_midY[i_sq4]])

    # Write numbers on image
    font                   = cv2.FONT_HERSHEY_SIMPLEX
    fontScale              = 5
    fontColor              = (0,255,0)
    lineType               = 10
    cv2.putText(img,'1', (sq1[0], sq1[1]), font, fontScale, fontColor, lineType)
    cv2.putText(img,'2', (sq2[0], sq2[1]), font, fontScale, fontColor, lineType)
    cv2.putText(img,'3', (sq3[0], sq3[1]), font, fontScale, fontColor, lineType)
    cv2.putText(img,'4', (sq4[0], sq4[1]), font, fontScale, fontColor, lineType)

    src = np.array([
        sq2,
        sq4,
        sq1,
        sq3], dtype="float32")
    w = width
    h = height
    if w>h:
        w=int(h/A4_SCALE)
    else:
        h=int(w*A4_SCALE)
    dst = np.array([
        [0, 0],
        [w, 0],
        [w, h],
        [0, h]], dtype = "float32")

    if debug:
        cv2.imwrite('temp.jpg', img)

    transform = cv2.getPerspectiveTransform(src, dst)
    img = cv2.warpPerspective(img, transform, (width, height))
    img = img[0:h, 0:w]
    img_gray = cv2.warpPerspective(img_gray, transform, (width, height))
    img_gray = img_gray[0:h, 0:w]
    return img, img_gray




# OMR - Optical mark regognition
# Copyright (C) 2019 Joppe Blondel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import cv2
import numpy as np
import math
import argparse
import zxing
from os import remove

from .parser import *

def read(fi, debug):
    fo = 'out.jpg'

    # Load image and create thresholded image img_th
    img, img_gray = prepare_image(fi, debug)
    height, width = img.shape[:2]

    # Read barcodes from bottom
    bc1 = img_gray[int(height*0.92):height, int(1*width/3):int(width/2)]
    bc2 = img_gray[int(height*0.92):height, int(width/2):int(2*width/3)]
    cv2.imwrite('bc1.png', bc1)
    cv2.imwrite('bc2.png', bc2)

    reader = zxing.BarCodeReader()
    res1 = reader.decode('bc1.png').raw
    res2 = reader.decode('bc2.png').raw
    b1 = res1.split(';')
    b2 = res2.split(';')

    if debug==False:
        remove('bc1.png')
        remove('bc2.png')

    # Barcode 1 data
    b1_ver = int(b1[0])
    # Version must be 1
    if b1_ver!=1:
        exit(-1)
    b1_col = int(b1[1])
    b1_row = int(b1[2])
    b1_block_col = int(b1[3])
    b1_block_row = int(b1[4])
    b1_startx = float(b1[5])/100    # Start of first block
    b1_starty = float(b1[6])/100
    b1_pady = float(b1[7])/100      # Space between blocks
    b1_padx = float(b1[8])/100

    b2_sizex = float(b2[0])/100     # Size of bullet
    b2_sizey = float(b2[1])/100
    b2_padx = float(b2[2])/100      # Space between bullets
    b2_pady = float(b2[3])/100

    mask = np.zeros(img.shape[:2], np.uint8)
    for col in range(0, b1_col):
        py = int(b1_startx*width + b1_pady*width*col)
        sy = int(b2_padx*b1_block_col*width)
        for row in range(0, b1_row):
            px = int(b1_starty*height + b1_padx*height*row)
            sx = int(b2_pady*b1_block_row*height)

            cv2.rectangle(img, (py, px), (py+sy, px+sx), (0, 0, 255), 3)

            for block_row in range(0, b1_block_row):
                bpx = px + int(b2_pady*height*block_row)
                sbx = int(b2_sizey*height)
                sby = int(b2_sizex*width)

                colors = [0] * b1_block_col
                for block_col in range(0, b1_block_col):
                    bpy = py + int(b2_padx*width*block_col)

                    cv2.rectangle(img, (bpy, bpx), (bpy+sby, bpx+sbx), (180, 180, 0), 2)
                    cv2.rectangle(mask, (bpy, bpx), (bpy+sby, bpx+sbx), (255, 255, 255), -1)
                    colors[block_col] = cv2.mean(img_gray, mask)[0]
                    cv2.rectangle(mask, (bpy, bpx), (bpy+sby, bpx+sbx), (0, 0, 0), -1)

                ans = np.argmin(colors)
                bpy = py + int(b2_padx*width*ans)
                cv2.rectangle(img, (bpy, bpx), (bpy+sby, bpx+sbx), (255, 0, 255), 4)
                print(ans, end=',')

    print("")
    if debug:
        cv2.imwrite(fo, img)




# OMR - Optical mark regognition
# Copyright (C) 2019 Joppe Blondel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import cv2
import numpy as np
import argparse
import urllib.request
from os import remove

from .parser import *

def create(fi, fo, col, row, bcol, brow, debug):

    transform, doc, _ = prepare_image_create(fi)
    if debug:
        cv2.imwrite('transformed.png', doc)
    height, width = doc.shape[:2]

    # Create red and green only masks
    (B, G, R) = cv2.split(doc)
    start = R-B
    end = G-B
    doc = cv2.merge([R+G-B, R+G-B, R+G-B])

    # Delete green and red squares for output
    (B, G, R) = cv2.split(cv2.imread(fi))
    out = cv2.merge([R+G-B, R+G-B, R+G-B])

    # Find coords of start and end squares
    start_c, start_h = cv2.findContours(start, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    end_c, end_h = cv2.findContours(end, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

    block_start = []
    block_end = []

    # GIVEN
    b1 = [""]*9
    b1_ver = 1
    b1_col = col
    b1_row = row
    b1_block_col = bcol
    b1_block_row = brow
    b1_startx = 0.0
    b1_starty = 0.0
    b1_padx = 0.0
    b1_pady = 0.0

    b2 = [""]*4
    b2_sizex = 0.0
    b2_sizey = 0.0
    b2_padx = 0.0
    b2_pady = 0.0

    for i in range(0, len(start_c)):
        # If there is no parent
        if start_h[0][i][3]==-1:
            block_start.append(i)
            block_end.append(i)

            p_s = cv2.arcLength(start_c[i], True)
            a_s = cv2.approxPolyDP(start_c[i], 0.03* p_s, True)
            p_e = cv2.arcLength(end_c[i], True)
            a_e = cv2.approxPolyDP(end_c[i], 0.03* p_s, True)

            # Must be rectangles
            if len(a_s)!=4 or len(a_e)!=4:
                exit()

            # Find bullet size if not yet found
            if b2_sizex==0 or b2_sizey==0:
                c1 = a_s[0][0]
                c2 = a_s[2][0]
                c3 = a_s[1][0]
                b2_sizey = (float(np.linalg.norm(c1-c3))/height)*100
                b2_sizex = (float(np.linalg.norm(c2-c3))/width)*100

            # Find space between bullets if not yet found
            if b2_padx==0 or b2_pady==0:
                # TODO not neat
                b2_padx = (float(a_e[0][0][0]+3 - a_s[0][0][0])/(width*(b1_block_col-1)))*100
                b2_pady = (float(a_e[0][0][1]+1 - a_s[0][0][1])/(height*(b1_block_row-1)))*100

            cv2.drawContours(doc, [start_c[i]], 0, (0, 0, 255), 4)
            cv2.drawContours(doc, [end_c[i]], 0, (0, 255, 0), 4)

            sx = start_c[i][0][0][0]
            sy = start_c[i][0][0][1]
            ex = end_c[i][0][0][0]
            ey = end_c[i][0][0][1]
            cv2.putText(doc, str(i), (sx, sy), cv2.FONT_HERSHEY_SIMPLEX, 3, (0, 0, 255), 5)
            cv2.putText(doc, str(i), (ex, ey), cv2.FONT_HERSHEY_SIMPLEX, 3, (0, 255, 0), 5)

    # calculate start position
    bstart = block_start[-1:][0]
    p_b = cv2.arcLength(start_c[bstart], True)
    a_b = cv2.approxPolyDP(start_c[bstart], 0.03* p_s, True)
    startx = a_b[0][0][0]
    starty = a_b[0][0][1]
    b1_startx = (float(startx)/width)*100
    b1_starty = (float(starty)/height)*100

    bnext = block_start[-2:][0]
    p_b = cv2.arcLength(start_c[bnext], True)
    a_b = cv2.approxPolyDP(start_c[bnext], 0.03* p_s, True)
    nextx = a_b[0][0][0]
    nexty = a_b[0][0][1]

    bdown = block_start[-1*b1_col-1:][0]
    p_b = cv2.arcLength(start_c[bdown], True)
    a_b = cv2.approxPolyDP(start_c[bdown], 0.03* p_s, True)
    downx = a_b[0][0][0]
    downy = a_b[0][0][1]

    b1_pady = ((downy - starty)/height)*100
    b1_padx = ((nextx - startx)/width)*100

    b1[0] = str(b1_ver)
    b1[1] = str(b1_col)
    b1[2] = str(b1_row)
    b1[3] = str(b1_block_col)
    b1[4] = str(b1_block_row)
    b1[5] = "{:10.3f}".format(b1_startx)
    b1[6] = "{:10.3f}".format(b1_starty)
    b1[7] = "{:10.3f}".format(b1_padx)
    b1[8] = "{:10.3f}".format(b1_pady)

    b2[0] = "{:10.3f}".format(b2_sizex)
    b2[1] = "{:10.3f}".format(b2_sizey)
    b2[2] = "{:10.3f}".format(b2_padx)
    b2[3] = "{:10.3f}".format(b2_pady)

    bc1 = ';'.join(b1).replace(" ", "")
    bc2 = ';'.join(b2).replace(" ", "")

    if debug:
        cv2.imwrite('doc.png', doc)

    # Create barcodes
    bc1 = bc1.replace(';','%3B')
    bc2 = bc2.replace(';','%3B')

    urllib.request.urlretrieve('http://datamatrix.kaywa.com/img.php?s=6&d=' + bc1, 'bc1.png')
    urllib.request.urlretrieve('http://datamatrix.kaywa.com/img.php?s=6&d=' + bc2, 'bc2.png')
    bc1 = cv2.imread('bc1.png')
    bc1_height, bc1_width = bc1.shape[:2]
    bc2 = cv2.imread('bc2.png')
    bc2_height, bc2_width = bc2.shape[:2]

    if debug==False:
        remove('bc1.png')
        remove('bc2.png')

    # Copy new barcodes over old ones
    oheight, owidth = out.shape[:2]
    codes = np.zeros((height, width, 3), dtype="uint8")

    h = int(0.045*height)
    w = int(h)
    sy = int(0.941*height)
    sx = int(0.396*width)
    # copy bc1 to here
    bc1 = cv2.resize(bc1, (h, w), interpolation = cv2.INTER_AREA)
    codes[sy:sy+h, sx:sx+h] = 255-bc1

    sx = int((1-0.396-0.045)*width)
    # copy bc2 to here
    bc2 = cv2.resize(bc2, (h, w), interpolation = cv2.INTER_AREA)
    codes[sy:sy+h, sx:sx+h] = 255-bc2

    invtransform = np.linalg.inv(transform)
    codes = cv2.warpPerspective(codes, invtransform, (owidth, oheight))

    cv2.imwrite(fo, cv2.bitwise_and(out, 255-codes))



from setuptools import setup

with open("README.md", 'r') as f:
    long_description = f.read()

setup(
    name='qandaomr',
    version='1.2.3',
    description='Q&A Python and OpenCV OMR (optical mark recognition)',

    author='Joppe Blondel',
    author_email='joppe@blondel.nl',
    download_url='https://github.com/Jojojoppe/qandaomr/archive/v1.2.3.tar.gz',
    url='https://github.com/Jojojoppe/qandaomr',

    keywords = ['OMR', 'Optical Mark Recognition',],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
  ],

    packages=['qandaomr'],
    licence='GNU General Public License v3 (GPLv3)',
    install_requires=['opencv-python', 'numpy', 'argparse', 'zxing'],
    scripts=['scripts/qandaomr']
)
